#include "4Gewinnt.h"

_4Gewinnt::_4Gewinnt()
{
    Clear();
}

void _4Gewinnt::Clear()
{
    for(int x = 0; x < FIELD_W; x++)
        for(int y = 0; y < FIELD_H; y++)
            field[x][y] = 0;
    player = 1;
}

int _4Gewinnt::Move(int move)
{
    int temp = -1;
    if((temp = NextHole(move)) != -1)   //Ist die Spalte schon voll?
    {                               
        field[move][temp] = player;     //Zug ausfhren
        player = -player;            //Anderer Spieler ist dran
    }
    return temp;        
}
int _4Gewinnt::Undo(int move)
{
    int temp;
    if((temp = NextHole(move)) == 0)    //Ohne Spielstein in der Spalte kann
        return -1;                      //auch kein Zug zurckgenommen werden
    else
    {
        if(temp == -1)
            temp = FIELD_H;
        field[move][temp - 1] = 0;      //Zug zurcknehmen
        player = -player;            //Anderer Spieler ist dran
    }
}
int _4Gewinnt::NextHole(int column)
{
    for(int i = 0; i < FIELD_H; i++)    //Spalte von unten nachoben durchzhlen
        if(field[column][i] == 0)       //Ist das Feld leer? 
            return i;
    return -1;                          //Spalte voll - kein leeres Feld
}

unsigned long _4Gewinnt::GetHashCode()
{
    unsigned long hash = 0;
    for(int x = 0, multi = 1; x < FIELD_W; x++, multi *= FAKTOR)
    {
        for(int y = 0, z = 1; y < FIELD_H; y++, z *= 2)
            hash += ((field[x][y] + 1) / 2) * z  * multi;
    }
    return hash;
}
  
int _4Gewinnt::DoMove()
{
    Move(Min(6,6));
    #ifdef HASHTABLE
    hash.clear();
    #endif
}

int _4Gewinnt::Max(int depth, int maxdepth)
{
    if(depth == 0)
        return Rate();
    int best = -10000;
    int move = -1;
    for(int i = 0; i < FIELD_W; i++)
        if(NextHole(i) != -1)
        {
            Move(i);
            int temp;
            if(ABS(temp = Rate()) == 1000)
            {
                Undo(i);
                if(depth == maxdepth)
                    return i;
                return temp;
            }
            int ist = Min(depth - 1, maxdepth);
            if(ist >= best)
            {

                move = i;
                best = ist;
            }
            
           
            Undo(i); 
        }
    if(depth == maxdepth)
        return move;
    else
        return best;
}
int _4Gewinnt::Min(int depth, int maxdepth)
{
    if(depth == 0)
        return Rate();
    int best = 10000;
    int move = -1;
    for(int i = 0; i < FIELD_W; i++)
        if(NextHole(i) != -1)
        {
            Move(i);
            int temp;
            if(ABS(temp = Rate()) == 1000)
            {
                Undo(i);
                if(depth == maxdepth)
                    return i;
                return temp;
            }
            int ist = Max(depth - 1, maxdepth);
            if(ist <= best)
            {
                move = i;
                best = ist;
            }
            Undo(i); 
        }
    if(depth == maxdepth)
        return move;
    else
        return best;
}


int _4Gewinnt::Rate()
{
    #ifdef HASHTABLE
    unsigned long code = GetHashCode();
    if(hash.find(code) != hash.end())
        return hash[code];
    #endif
    int result = 0;
    for(int x = 0; x < FIELD_W; ++x)
        for(int y = 0; y < FIELD_H; ++y)
        {
            result += field[x][y] * (3 - ABS(x - 3)/* * 3 / (FIELD_W / 2)*/);
            for(int dx = -1; dx < 1; ++dx)          //Von -1 zu (!) 0
                for(int dy = -1; dy < 2; ++dy)
                {
                    if((dx || dy) && (dx + dy) != 1 && field[x][y])     //0-0 wird gestrichen
                    {         
                        int run = 0;
                        int free[] = {0,0};
                        int edx, edy;
                        for(int i = -1; i < 2; i += 2 )
                        {
                            edx = dx * i;
                            edy = dy * i;
                            while(InField(x+edx,y+edy))
                            {
                                if(field[x][y] == field[x+edx][y+edy] && !free[(i + 1) / 2])
                                    run++;
                                else if(field[x+edx][y+edy] == 0)
                                    free[(i + 1) / 2]++;
                                else
                                    break;
                                edx += dx * i;
                                edy += dy * i;
                            }
                        }
                        if(run >= 3)
                            return field[x][y] * 1000;
                        if(run > 0 && run + free[0] + free[1] >= 3)
                        {
                            result += field[x][y]*(run*run + free[0] + free[1]);
                        }
                    }
                }
        }
    #ifdef HASHTABLE
    hash[code] = result;
    #endif
    return result;
}
